/*--------------------------------------------------------------------------------*/
/* Cstartup.s                                                                     */
/* Asemblerovsky startovaci subor pre ARM                                         */
/*--------------------------------------------------------------------------------*/

/* Definicia velkosti zasobnika */
     .equ   IRQ_Stack_Size,     0x00000060

     .equ AIC_IVR,         (256) 
     .equ AIC_FVR,         (260) 
     .equ AIC_EOICR,       (304)
     .equ AT91C_BASE_AIC,  (0xFFFFF000)

/*--------------------------------------------------------------------------------*/
/* Definicia sekcii                                                               */
/*-----------------                                                               */
/* Sekcia                                                                         */    
/* .internal_ram_top 	Top_Stack: pouzity v cstartup pre inicializaciu           */
/*                      vektorov definovanych linkerom a ovplyvnenych ld skriptom */
/*--------------------------------------------------------------------------------*/
	.section 	.internal_ram_top
	.code 32
	.align 	0
	.global	Top_Stack
Top_Stack:
	
/*--------------------------------------------------------------------------------*/
/* Definicia oblasti                                                              */
/*--------------------------------------------------------------------------------*/
        .section 	.reset
        .text
        .global _startup
        .func   _startup

_startup:
reset: 
/*--------------------------------------------------------------------------------*/
/* Vektory vynimiek                                                               */
/*--------------------------------------------------------------------------------*/
/* Tieto vektory mozu byt precitane na adrese 0 alebo na RAM adrese               */
/* Nutne musia byt v relativnom adresovacom mode, aby bol garantovany platny skok */
/* Momentalne obsahuju len nekonecnu slucku.                                      */
/* Ak nastane vynimka pred remapovanim, dojde k zacykleniu.                       */
/*--------------------------------------------------------------------------------*/

                B           InitReset           /* 0x00 Obsluha Resetu */
undefvec:
                B           undefvec            /* 0x04 Nedefinovana instrukcia */
swivec:
                B           swivec              /* 0x08 Softverove prerusenie */
pabtvec:
                B           pabtvec             /* 0x0C Prefetch Abort */
dabtvec:
                B           dabtvec             /* 0x10 Data Abort */
rsvdvec:
                B           rsvdvec             /* 0x14 rezervovane */
irqvec:
                B           IRQ_Handler_Entry   /* 0x18 IRQ */
fiqvec:               	            	        /* 0x1c FIQ */
/*--------------------------------------------------------------------------------*/
/* Funkcia FIQ_Handler_Entry                                                      */
/* Sluzi na obsluhu FIQ preruseni                                                 */
/* Volaju sa funkcie AIC_FVR[interrupt]                                           */
/*--------------------------------------------------------------------------------*/

FIQ_Handler_Entry:

/* Prepnutie do SVC/User modu na povolenie pristupu do zasobnika pre C program */
/* kedze FIQ este nie je potvrdeny */

/* Ulozenie r0 do FIQ_Register */
            mov         r9,r0
            ldr         r0 , [r8, #AIC_FVR]
            msr         CPSR_c,#I_BIT | F_BIT | ARM_MODE_SVC

/* Odlozenie pouzitych registrov a LR do uzivatelskeho zasobnika */
            stmfd       sp!, { r1-r3, r12, lr}

/* Skok na obsluhu podla AIC_FVR */
            mov         r14, pc
            bx          r0

/* Vratenie registrov a LR zo zasobnika */
            ldmia       sp!, { r1-r3, r12, lr}

/* Prerusenia ostavaju zakazane a prepnutie do FIQ modu */
            msr         CPSR_c, #I_BIT | F_BIT | ARM_MODE_FIQ

/* Vratenie R0 ARM_MODE_SVC register */
            mov         r0,r9

/* Vratenie Program Countera pouzitim LR_fiq priamo do PC */
            subs        pc,lr,#4
	.align 0
.RAM_TOP:
	.word	Top_Stack

InitReset:
/*--------------------------------------------------------------------------------*/
/* Low level inicializacia (PMC, AIC, ? ....) C funkciou AT91F_LowLevelInit       */
/*--------------------------------------------------------------------------------*/
            	.extern   AT91F_LowLevelInit
/* Volanie AT91F_LowLevelInit(void) */

            ldr     r13,.RAM_TOP            /* docasny zasobnik vo vnutornej RAM */
/* Volanie funkcie AT91F_LowLevelInit cez Interworking */
            ldr     r0,=AT91F_LowLevelInit
            mov     lr, pc
            bx      r0

/*--------------------------------------------------------------------------------*/
/* Definicia velkosti zasobnikov                                                  */
/*--------------------------------------------------------------------------------*/
/* Zasobnik preruseni vyzaduje 2 slova x 8 urovni priority x 4 bajty, ked sa      */
/* pouzivaju vektory.                                                             */
/* Zasobnik preruseni musi byt upraveny podla obsluhy preruseni.                  */
/* Rychle prerusenia (FIQ) nepotrebuju zasobnik. Ak ho vyzaduje aplikacia,        */
/* musi byt definovany inde.                                                      */
/* Systemovy zasobnik nie je definovany a je obmedzeny velkostou volnej vnutornej */
/* SRAM.                                                                          */
/*--------------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------------*/
/* Definicia vrcholu zasobnika                                                    */
/*--------------------------------------------------------------------------------*/
/* Zasobnik preruseni a Supervisor zasobnik su umiestnene na vrchu internej       */
/* pamate, kvoli zrychleniu ukladania a obnovovania kontextu.                     */
/* Zasobnik ARM_MODE_SVC (Aplikacia, C) je umistneny na vrchu externej pamate     */
/*--------------------------------------------------------------------------------*/

	  .EQU		IRQ_STACK_SIZE,    (3*8*4)
          .EQU		ARM_MODE_FIQ,       0x11
          .EQU		ARM_MODE_IRQ,       0x12
          .EQU		ARM_MODE_SVC,       0x13

          .EQU		I_BIT,              0x80
          .EQU		F_BIT,              0x40

/*--------------------------------------------------------------------------------*/
/* Nastavenie zasobnika pre kazdy mod                                             */
/*--------------------------------------------------------------------------------*/
                mov     r0,r13

/* Nastavenie modu Fast Interrupt a nastavenie FIQ zasobnika */
                msr     CPSR_c, #ARM_MODE_FIQ | I_BIT | F_BIT
/* Inicializacia FIQ registra */
            	ldr     r8, =AT91C_BASE_AIC

/* Nastavenie modu Interrupt a nastavenie IRQ zasobnika */
                msr     CPSR_c, #ARM_MODE_IRQ | I_BIT | F_BIT
                mov     r13, r0                     /* Inicializacia zasobnika IRQ */
                sub     r0, r0, #IRQ_Stack_Size
/* Nastavenie modu Supervisor a nastavenie zasobnika */
                msr     CPSR_c, #ARM_MODE_SVC
                mov     r13, r0                     /* Inicializacia zasobnika Sup */

/* Realokovanie .data sekcie (kopirovanie z ROM do RAM */
                LDR     R1, =_etext
                LDR     R2, =_data
                LDR     R3, =_edata
LoopRel:        CMP     R2, R3
                LDRLO   R0, [R1], #4
                STRLO   R0, [R2], #4
                BLO     LoopRel

/* Vymazanie .bss sekcie (inicializacia nulami) */
                MOV     R0, #0
                LDR     R1, =__bss_start__
                LDR     R2, =__bss_end__
LoopZI:         CMP     R1, R2
                STRLO   R0, [R1], #4
                BLO     LoopZI

		ldr	lr,=exit
		ldr	r0,=main
		bx	r0
		
        .size   _startup, . - _startup
        .endfunc
		
/* Prazdna funkcia "exit", aby nebolo potrebne sbrk write read atd., co je */
/* potrebne v defaultnej funkcii "exit" v newlib */
        .global exit
        .func   exit
exit:
        b    .
		.size   exit, . - exit
        .endfunc
		
/*--------------------------------------------------------------------------------*/
/* Obsluha vynimiek                                                               */
/*--------------------------------------------------------------------------------*/
/* Tato cast musi byt v ARM mode                                                  */
/*--------------------------------------------------------------------------------*/
/* Funkcia IRQ_Handler_Entry                                                      */
/* Sluzi na obsluhu IRQ preruseni                                                 */
/* Volaju sa funkcie AIC_IVR[interrupt]                                           */
/*--------------------------------------------------------------------------------*/
        .global IRQ_Handler_Entry
        .func   IRQ_Handler_Entry

IRQ_Handler_Entry:

/* Uprava a ulozenie LR_irq do IRQ zasobnika */
            sub         lr, lr, #4
            stmfd       sp!, {lr}

/* Ulozenie SPSR kvoli vnorenym preruseniam */
            mrs         r14, SPSR
            stmfd       sp!, {r14}

/* Ulozenie r0 do IRQ zasobnika */
            stmfd       sp!, {r0}

/* Zapis do IVR na podporu Protect modu */
/* Nema vplyv na normalny mod */
/* Zrusenie NIRQ a vymazanie zdroja v Protect mode */
            ldr         r14, =AT91C_BASE_AIC
	    ldr         r0 , [r14, #AIC_IVR]
	    str         r14, [r14, #AIC_IVR]

/* Povolenie preruseni a prepnutie do Supervisor modu */
            msr         CPSR_c, #ARM_MODE_SVC

/* Ulozenie pouzitych registrov a LR do uzivatelskeho zasobnika */
            stmfd       sp!, { r1-r3, r12, r14}

/* Skok na obsluhu podla AIC_IVR  */
            mov         r14, pc
            bx          r0
/* Vratenie registrov a LR zo zasobnika */
            ldmia       sp!, { r1-r3, r12, r14}

/* Zakazanie preruseni a prepnutie do IRQ modu */
            msr         CPSR_c, #I_BIT | ARM_MODE_IRQ

/* Oznacenie konca prerusenia v AIC */
            ldr         r14, =AT91C_BASE_AIC
            str         r14, [r14, #AIC_EOICR]

/* Vratenie SPSR_irq a r0 z IRQ zasobnika */
            ldmia       sp!, {r0}

/* Vratenie SPSR_irq a r0 z IRQ zasobnika */
            ldmia       sp!, {r14}
            msr         SPSR_cxsf, r14

/* Vratenie upraveneho LR_irq z IRQ zasobnika priamo do PC */
            ldmia       sp!, {pc}^
	
        .size   IRQ_Handler_Entry, . - IRQ_Handler_Entry
        .endfunc

/*--------------------------------------------------------------------------------*/
/* ?Vektory vynimiek                                                              */
/* Tato cast je prilinkovana, len ak je to potrebne.                              */
/*--------------------------------------------------------------------------------*/
        .global AT91F_Default_FIQ_handler
        .func   AT91F_Default_FIQ_handler
AT91F_Default_FIQ_handler:
            b     AT91F_Default_FIQ_handler
        .size   AT91F_Default_FIQ_handler, . - AT91F_Default_FIQ_handler
        .endfunc

        .global AT91F_Default_IRQ_handler
        .func   AT91F_Default_IRQ_handler
AT91F_Default_IRQ_handler:
            b     AT91F_Default_IRQ_handler
        .size   AT91F_Default_IRQ_handler, . - AT91F_Default_IRQ_handler
        .endfunc

        .global AT91F_Spurious_handler
        .func   AT91F_Spurious_handler
AT91F_Spurious_handler:
            b     AT91F_Spurious_handler
        .size   AT91F_Spurious_handler, . - AT91F_Spurious_handler
        .endfunc

        .end
